/****************************************************************************
*****************************************************************************
**   The xGlobal extender                                                  **
**                                                                         **
**   This DLL can be used to keep global variables for user defined        **
**   functions.    (All variables in a UDF are deleted when the            **
**   function exits, so without this extender there is no way to           **
**   keep any values from run to run.)                                     **
**                                                                         **
**   The defined routines are...                                           ** 
**                                                                         **
**     xGlobalPut(Name,Value)                                              **
**                                                                         **
**            "Name" is any non-blank string.  This plays the role of      **
**            a variable name.  "Value" is any value to be saved for       **
**            later recall.  If the "Variable name" has been defined       **
**            before, the value replaces it's previous value.              **
**            The maximum size of a name is 64 characters.                 **
**                                                                         **
**     xGlobalGet(Name,Value)                                              **
**                                                                         **
**            If you called "xGlobalPut" with this name, and a value,      **
**            xGlobalGet will return the value.  If the name was not       **
**            defined, then the return value is an empty string.           **
**                                                                         **
**     xGlobalIsDefined(Name)                                              **
**                                                                         **
**            Returns TRUE (1) if the name has been defined by xGlobalPut. **
**            return FALSE (0) otherwise.                                  **
**                                                                         **
**      xGlobalDrop(Name)                                                  **
**                                                                         **
**            Deletes the saved name and value.                            **
**                                                                         **
**      xGlobalVersion(0)                                                  **
**                                                                         **
**            Returns the extender version number.                         **
**                                                                         **
*****************************************************************************
*****************************************************************************
**                                                                         **
**  By: Alan Kreutzer  --      kreutzer@ost.state.or.us                    **
**  04/2001                                                                **
**                                                                         **
**  Oregon State Treasury                                                  **
**  350 Winter St. Suite 100                                               **
**  Salem, OR  97301-3896                                                  **
**                                                                         **
**                                                                         **
**  Requires WinBatch version 2001 or later                                **
**                                                                         **
**  (c) 2001, Oregon State Treasury                                        **
**  May be freely distributed, but may not be sold.                        **
**  This program is distributed with no warrenty, use at your own risk.    **
**  The Oregon State Treasury assumes no responsibility for problems       **
**  caused by the use of this software.                                    ** 
**                                                                         **
*****************************************************************************
****************************************************************************/
#define EXTENDERVERSION 20100       // DLL version number.

#define INTERFACELEVEL  2           // extender interface level 
#define EXTENDER_FEATURE_LEVEL 1001 // extender feature level 

#define STRICT


/*---------------------------------------------------------------------------+
!                                   Includes                                 !
+---------------------------------------------------------------------------*/
#include <windows.h>
                           /* special includes for WinBatch extenders */
#include "wwwpform.h"
#include "addons.h"

/*---------------------------------------------------------------------------+
!                Debug Functions                                             !
+---------------------------------------------------------------------------*/
#define MBSS(p1, p2) {MessageBox(GetFocus(), (p2), (p1), MB_OK | MB_SETFOREGROUND);}
#define MBSN(p1, p2) {char szDebug[20]; wsprintf(szDebug,"%d",(long)(p2)); MessageBox(GetFocus(), szDebug, (p1), MB_OK | MB_SETFOREGROUND);}

/*---------------------------------------------------------------------------+
!       define the structure we use when we allocate                         !
!       a block of memory to hold a name and value.                          ! 
+---------------------------------------------------------------------------*/
typedef struct GlobalBlockStruct      // This structure defines the
{                                     // the block of memory we allocate
     struct GlobalBlockStruct *Chain; // to hold a name and value.
     long   NameLen;
     char   Text[1];
} GlobalBlock;

/*---------------------------------------------------------------------------+
!      RETLONG, RETFLOAT and RETSTRING return values to WinBatch,            !
!      and exit the extender.                                                !
+---------------------------------------------------------------------------*/
#define RETLONG(numb)         return (lpViperCallback(((lpViperReturn->x=numb)      +(lpViperReturn->inittype=VARTYPE_INT))      ?RESULT_SUCCESS:RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn))
#define RETFLOAT(numb)        return (lpViperCallback(((lpViperReturn->flt=numb)    +(lpViperReturn->inittype=VARTYPE_FLOATNUM)) ?RESULT_SUCCESS:RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn))
#define RETSTRING(str)        return (lpViperCallback(((lpViperReturn->lpstr=str)   +(lpViperReturn->inittype=VARTYPE_STRING))   ?RESULT_SUCCESS:RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn))

/*---------------------------------------------------------------------------+
!     ANSWERSTRING, ANSWERLONG, and ANSWERFLOAT pass values to WinBatch,     !
!     but do not exit the extender.  This allows you to do cleanup (such     !
!     as releasing allocated memory) before exit.  The return value from     !
!     the ANSWERSTRING function should be the extender function's return     !
!     value when it exits.                                                   !
+---------------------------------------------------------------------------*/
#define ANSWERSTRING(str) lpViperCallback(  ((lpViperReturn->lpstr=str)+(lpViperReturn->inittype=VARTYPE_STRING))?RESULT_SUCCESS:RESULT_SUCCESS,ViperInterfaceVersion,lpViperStruct,lpViperReturn)
#define ANSWERLONG(numb)  lpViperCallback(  ((lpViperReturn->x=numb)+(lpViperReturn->inittype=VARTYPE_INT))?RESULT_SUCCESS:RESULT_SUCCESS,ViperInterfaceVersion,lpViperStruct,lpViperReturn)
#define ANSWERFLOAT(numb) lpViperCallback(  ((lpViperReturn->flt=numb)+(lpViperReturn->inittype=VARTYPE_FLOATNUM))?RESULT_SUCCESS:RESULT_SUCCESS,ViperInterfaceVersion,lpViperStruct,lpViperReturn)

/*---------------------------------------------------------------------------+
! RETERROR(level,rcfilestringid) returns an error massage to WinBatch        !
!   level MUST be -1, -2 or -3.                                              ! 
!     Use -1 for most run of the mill errors.                                !
!         -2 for serious errors                                              !
!         -3 for absolutely fatal "we better die" errors                     !
!   rcstringid is the string ID from the RC file that contains the error     !
!   message.  The first three characters of the string should be the same    ! 
!   as the string ID, and the fourth character should be a colon.            !
+---------------------------------------------------------------------------*/
#define RETERROR(sev, errnum) return (lpViperCallback(((LoadString(hinstDLL, (errnum), lpViperReturn->lpstr = szWork, sizeof(szWork)))+(lpViperReturn->inittype = VARTYPE_STRING)) ?(sev)         :(sev)         , ViperInterfaceVersion, lpViperStruct, lpViperReturn))

/*---------------------------------------------------------------------------+
!      RETCANCEL returns a Cancel flag                                       !
+---------------------------------------------------------------------------*/
#define RETCANCEL return (-5)




/*---------------------------------------------------------------------------+
!                           Function Prototypes                              !
+---------------------------------------------------------------------------*/
long CopyLower(char *Buffer, char *String);

/*---------------------------------------------------------------------------+
!                                Global Variables!
+---------------------------------------------------------------------------*/
                 /*- Superstitious data field to bypass/ignore bugs in
                 !   certain versions of Windows that cause the beginning 
                 !   of the data area to be overwritten.  This field should 
                 !   be the first data area defined and linked. 
                 +-*/
char szBugFiller[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

HINSTANCE hinstDLL;                    // DLL Instance saved


/*--------------------------------------------------------------------------+
!  Command Extension Table for this DLL                                     !
+--------------------------------------------------------------------------*/
tablestruct CommandExtTable[] =
{
//----------------------------------------------------------------------
// ID   Type       Mask              Params/Value, Size, Name
//----------------------------------------------------------------------
  {10,  FUNCTION, 0x0000000000000000, 0x00000001L, 14, "xGlobalVersion"  }, 
  {11,  FUNCTION, 0x0000000000000011, 0x00000002L, 10, "xGlobalPut"      }, 
  {12,  FUNCTION, 0x0000000000000001, 0x00000001L, 10, "xGlobalGet"      },
  {13,  FUNCTION, 0x0000000000000001, 0x00000001L, 16, "xGlobalIsDefined"},
  {14,  FUNCTION, 0x0000000000000001, 0x00000001L, 11, "xGlobalDrop"     }, 
//--------------------------------------------------------------------------
  {-1,         0,                   0,  0x00000000,  0, ""             }
};

/*******************************************************************************
********************************************************************************
**                                                                            **
**                                    DllMain                                 **
**                                                                            **
********************************************************************************
*******************************************************************************/
BOOL WINAPI DllMain(HANDLE hInst, DWORD dwReason, LPVOID lpReserved)
{

  hinstDLL = hInst;
  UNREFERENCED_PARAMETER(dwReason);
  UNREFERENCED_PARAMETER(lpReserved);
  return (TRUE);
}

/*******************************************************************************
********************************************************************************
**                                                                            **
**                              WILExtenderQuery                              **
**                                                                            **
**           Tells the WilDll about our interface level.                      **
**                                                                            **
********************************************************************************
*******************************************************************************/
DllExport LONG FAR PASCAL WILExtenderQuery
 (short  nRequest,                     // type of information being requested
  LPVOID lpvInfo,                      // information buffer
  LONG   nCount)                       // additional data
{
  char szWork[256];
  LONG        i;

  switch (nRequest)
  {
        /*--------------------------------------------------------------+
        !  case 1 ... identify yourself to a caller.                    !
        +--------------------------------------------------------------*/
    case 1:                      
    {
      if ((lpvInfo) && (nCount > 0))
      {
        LoadString(hinstDLL, 1000, szWork, sizeof(szWork));
        lstrcpyn((LPSTR) lpvInfo, szWork, nCount);
      }
      return (EXTENDERVERSION);
    }

        /*--------------------------------------------------------------+
        !  case 2 ... Tell WinBatch we use interface level 2            !  
        +--------------------------------------------------------------*/
    case 2:                            // preferred interface version
    {
      for (i = 0; i < nCount; i++)
        if (((LPLONG) lpvInfo)[i] == INTERFACELEVEL)  // found one we like
          return (INTERFACELEVEL);

      return (-1);                     // didn't find any we like
    }


        /*--------------------------------------------------------------+
        !  unknown requests                                             ! 
        +--------------------------------------------------------------*/
    default:
      return (0);
  }
}

/*******************************************************************************
********************************************************************************
**                                                                            **
**                          WILExtender2                                      **
**                                                                            **
**   The main routine performs all extender functions, and various            **
**   initialization and cleanup tasks.                                        **
**                                                                            **
********************************************************************************
*******************************************************************************/

DllExport LONG FAR PASCAL WILExtender2
  (HWND           hWnd,                   // hWnd of parent application
  HINSTANCE       hInst,                  // hInstance of parent application
  LONG            ViperInterfaceVersion,  // WIL interface version
  short           CmdId,                  // identifies extender function being called
  LPVIPERCALLBACK lpViperCallback,        // pointer to return value callback function
  LPVIPERSTRUCT   lpViperStruct,          // pointer to information structure
  LPLONG          lpLong,                 // pointer to your 32 bits
  LPVIPERVAR      lpvvArg,                // pointer to parameter array
  LONG            nArgCount,              // parameter count for lpvvArg
  LPVIPERVAR      lpViperReturn)          // pointer to return value structure
{     
  long L;
  GlobalBlock Root;                       // Some local storage
  GlobalBlock *Block;
  GlobalBlock *NewBlock;
  GlobalBlock *PreviousBlock;
  char szWork[256];          

             
  switch (CmdId)          // the "big switch" (cases < 0 are at the bottom)
  {
        /*--------------------------------------------------------------+
        !  case -1 ... One time initialization routine.                 !
        !              Return value must be number of functions         !
        !              & constants defined.                             !  
        +--------------------------------------------------------------*/
    case -1:                         
    {
       return ((sizeof(CommandExtTable) / sizeof(tablestruct)) - 1); 
    }

        /*--------------------------------------------------------------+
        !  case -2 ... Called repeatedly as extender is being loaded.   !
        !              On each call, return a function table entry.     !
        +--------------------------------------------------------------*/
    case -2:                          
    {
      LPTABLESTRUCT lpTable;

      lpTable = (LPTABLESTRUCT) lpvvArg;
      L = (UINT) nArgCount;

      if (CommandExtTable[L].ident == -1)
        return (0);

      if (lpTable)
      {
        lpTable->ident  = CommandExtTable[L].ident;
        lpTable->type   = CommandExtTable[L].type;
        lpTable->param1 = CommandExtTable[L].param1;
        lpTable->ulMask = CommandExtTable[L].ulMask;
        lpTable->size   = CommandExtTable[L].size;
        lstrcpy((LPSTR) lpTable->delim, (LPSTR) CommandExtTable[L].delim);
        CharLower(lpTable->delim);
      }

      return (1);
    }

        /*--------------------------------------------------------------+
        !  case -3 .. Called once at shutdown. Can be used for cleanup. !
        !             We use it to release all allocated memory blocks. !
        +--------------------------------------------------------------*/
    case -3:
    {
   
        Root.Chain = (GlobalBlock *) *lpLong;
        while (Root.Chain != NULL)
        {
             Block = Root.Chain;
             Root.Chain = Block->Chain;
             GlobalFree(Block);
        }
        *lpLong = (long) NULL;
        return (1);
    }

        /*--------------------------------------------------------------+
        !  case 10 ... Return extender version information.             !
        !              Parameter 0 = Version                            !
        !                        1 = Total Functions                    !
        !                        2 = Total Constants                    !
        +--------------------------------------------------------------*/
    case 10:
    {
         long Param;         
         UINT uCount, uType;

         Param = lpvvArg[0].x;
         if (Param == 0)
              RETLONG(EXTENDERVERSION);         
   
         if (Param == 1)                     // total functions
             uType = FUNCTION;
         else 
            if (Param == 2)                  // total constants
                uType = CONSTANT;
            else
                RETERROR(-1, 101);           // invalid parameter
      
        for (L = uCount = 0; CommandExtTable[L].ident != -1; L++)
           if ((UINT) CommandExtTable[L].type == uType)
             uCount++;

        RETLONG(uCount);
    }

        /*--------------------------------------------------------------+
        !  case 11 ... xGlobalPut(Name, Value)                          !
        !             Stores the name and value in our list.            !
        +--------------------------------------------------------------*/
    case 11:
    {
          long LName;
          long LValue; 
          char *pValue;    
                            //-----------------------------------------------
                            //  Convert the name to lower case and see how
                            //  long it is.  It can't be zero or > 64.
                            //-----------------------------------------------
          LName = CopyLower(szWork,lpvvArg[0].lpstr); 
          if (LName == 0) RETERROR(-1,102);          // Can't be totally blank.
          if (LName > 64) RETERROR(-1,103);          // Name too long.

                            // ----------------------------------------------
                            //  Measure the value.  Get a memory block 
                            //  long enough to hold the name and value,
                            //  and move both of them to the block.
                            // ----------------------------------------------                 
          LValue =lstrlen(lpvvArg[1].lpstr);        // Measure the value
          L = sizeof(GlobalBlock) + LName + LValue + 2;   
          NewBlock = (GlobalBlock *) GlobalAlloc(GMEM_FIXED,L);
	      if (NewBlock == NULL) RETERROR(-3,913);   // Couldn't allocate memory?

          NewBlock->Chain = NULL;
          NewBlock->NameLen = LName;
                                         
          lstrcpy(NewBlock->Text,szWork);         // Copy name into block.
          pValue = NewBlock->Text+LName+1;
          lstrcpy(pValue,lpvvArg[1].lpstr);       // Copy Value into block.

 
                            // ----------------------------------------------
                            //  Now walk through the existing blocks looking
                            //  for a previous definition of this name.
                            // ----------------------------------------------
          Root.Chain = (GlobalBlock *) *lpLong;
          Block = &Root;
          while (1)
          {
               PreviousBlock = Block;
               Block = Block->Chain;
               if (Block == NULL) break;
               if (NewBlock->NameLen != LName) continue;
               if (lstrcmp(NewBlock->Text, Block->Text)== 0) break;
          }
                            // ----------------------------------------------
                            // If there was a previous definition, replace it.
                            // otherwise, add the new value to the chain.
                            // ----------------------------------------------
          if (Block == NULL)
          {
              PreviousBlock->Chain = NewBlock;
          }
          else
          {
              NewBlock->Chain = Block->Chain;
              PreviousBlock->Chain = NewBlock;   
              GlobalFree(Block);
          }
           *lpLong = (long) Root.Chain;
           RETLONG (1);

    }

        /*--------------------------------------------------------------+
        !  case 12 ... xGlobalGet(Name)                                 !
        !             Return a value from  our list.                    !
        +--------------------------------------------------------------*/
    case 12:
    {         
         char *pValue;
                            // ----------------------------------------------
                            //  Convert the name to lower case
                            // ----------------------------------------------
         L = CopyLower(szWork, lpvvArg[0].lpstr);
         if (L == 0) RETSTRING("");
         if (L > 64) RETSTRING("");
         
                            // ----------------------------------------------
                            // Search for name in memory
                            // ----------------------------------------------
         Root.Chain = (GlobalBlock *) *lpLong;
         Block = &Root;
         while (1)
         {
              Block = Block->Chain;
              if (Block == NULL) RETSTRING("");
              if (Block->NameLen != L) continue;
              if (lstrcmp(Block->Text, szWork)== 0) 
              {
                  pValue = Block->Text + Block->NameLen + 1;
                  RETSTRING(pValue);
              }
         }
    }

        /*--------------------------------------------------------------+
        !  case 13 ... xGlobalIsDefined(Name)                           !
        !             Return TRUE if a name has been defined            !
        +--------------------------------------------------------------*/
    case 13:
    {
                            // ----------------------------------------------
                            //  Convert the name to lower case
                            // ----------------------------------------------         
         L = CopyLower(szWork, lpvvArg[0].lpstr);;
         if (L == 0) RETLONG(0);
         if (L > 64) RETLONG(0);
         
                            // ----------------------------------------------
                            // Search for name in memory
                            //  If we find it, return TRUE.
                            // ----------------------------------------------
         Root.Chain = (GlobalBlock *) *lpLong;
         Block = &Root;
         while (1)
         {
              Block = Block->Chain;
              if (Block == NULL)   RETLONG(0);
              if (Block->NameLen != L) continue;
              if (lstrcmp(Block->Text, szWork)== 0) RETLONG(1);
         }		
    }


        /*--------------------------------------------------------------+
        !  case 14 ... xGlobalDrop(Name)                                !
        !             remove a name/value from our table.               !
        +--------------------------------------------------------------*/
    case 14:
    {
                            // ----------------------------------------------
                            //  Convert the name to lower case
                            // ---------------------------------------------- 
         L = CopyLower(szWork, lpvvArg[0].lpstr);
         if (L == 0) RETLONG(0);
         if (L > 64) RETLONG(0);
         
                            // ----------------------------------------------
                            // Search for name in memory
                            //  If we find it, drop it.
                            // ----------------------------------------------
         Root.Chain = (GlobalBlock *) *lpLong;
         Block = &Root;
         while (1)
         {
              PreviousBlock = Block;
              Block = Block->Chain;
              if (Block == NULL)   RETLONG(0);
              if (Block->NameLen != L) continue;
              if (lstrcmp(Block->Text, szWork)== 0)
              {
                  PreviousBlock->Chain = Block->Chain;
                  GlobalFree(Block);
                  *lpLong = (long) Root.Chain;
                  RETLONG(1);
              }
         }
    }



        /*--------------------------------------------------------------+
        !  Default Case.  (We should never get here)                    !
        +--------------------------------------------------------------*/
    default:      
    {
      RETERROR(-1, 901);               // Unknown function
    }
 }                                      // end of "big switch
}

/********************************************************************************
*********************************************************************************
**                                                                             **
**  Copy a string to the buffer and convert it to lower case.                  **
**                                                                             **
**  Leading and trailing spaces are stripped. A maximum of 64 characters are   **
**  moved.  (The buffer must be at least 65  characters long, so there is      **
**  room for a trailing \0.)                                                   **
**                                                                             **
**  The return value is the actual length of the string.  Caller should check  **
**  to make sure it is in the valid range 1 - 64.                              **
*********************************************************************************
********************************************************************************/
long CopyLower(char *Buffer, char *String)
{
    char *A, *B;
    long Len, CopyLen;
    
    while (*String == ' ') String++;  // Strip leading spaces
   
    A = String;
    while (*A != '\0') A++;           // Find the end of the string
    while (1)
    {
        if (A == String) break;       // Strip Trailing Spaces.
        B = A - 1;        
        if (*B != ' ') break;
        A = B;
    }
    Len = A - String;                 // Length of string
    if (Len > 64) CopyLen = 65;
    else          CopyLen = Len+1;
    lstrcpyn(Buffer, String, CopyLen);// Copy string and add terminator.
    CharLower(Buffer);                // Convert to lower case.
    return Len;
}